# -*- coding: utf-8 -*-

# Built-in modules #
import sys, os, time, re, random, math, json
import getpass, hashlib, datetime, collections
from itertools import compress, product

# Third party modules #
import numpy, dateutil

# One liners #
flatter = lambda x: [item for sublist in x for item in sublist]

################################################################################
def ascii(text):
    """Make a safe, ASCII version a string. For instance for use on the web."""
    import unicodedata
    return unicodedata.normalize('NFKD', unicode(text)).encode('ASCII', 'ignore')

################################################################################
def sanitize(text):
    """Make an ultra-safe, ASCII version a string.
    For instance for use as a filename."""
    return "".join([c for c in text if re.match(r'\w', c)])

################################################################################
def bool_to_unicode(b):
    """Different possibilities for True: ☑️✔︎✓✅👍
       Different possibilities for False: ✕✖︎✗✘✖️❌⛔️❎👎"""
    if not isinstance(b, bool): b = bool(b)
    if b is True:  return u"✅"
    if b is False: return u"❎"

################################################################################
def all_combinations(items):
    """Generate all combinations of a given list of items."""
    return (set(compress(items,mask)) for mask in product(*[[0,1]]*len(items)))

################################################################################
def pad_with_whitespace(string, pad=None):
    """Given a multiline string, add whitespaces to every line
    so that every line has the same length."""
    if pad is None: pad = max(map(len, string.split('\n'))) + 1
    return '\n'.join(('{0: <%i}' % pad).format(line) for line in string.split('\n'))

###############################################################################
def mirror_lines(string):
    """Given a multiline string, return its reflection along a vertical axis.
    Can be useful for the visualization of text version of trees."""
    return '\n'.join(line[::-1] for line in string.split('\n'))

###############################################################################
def concatenate_by_line(first, second):
    """Zip two strings together, line wise"""
    return '\n'.join(x+y for x,y in zip(first.split('\n'), second.split('\n')))

################################################################################
def sort_string_by_pairs(strings):
    """Group a list of strings by pairs, by matching those with only
    one character difference between each other together."""
    assert len(strings) % 2 == 0
    pairs = []
    strings = list(strings) # This shallow copies the list
    while strings:
        template = strings.pop()
        for i, candidate in enumerate(strings):
            if count_string_diff(template, candidate) == 1:
                pair = [template, strings.pop(i)]
                pair.sort()
                pairs.append(pair)
                break
    return pairs

################################################################################
def count_string_diff(a,b):
    """Return the number of characters in two strings that don't exactly match"""
    shortest = min(len(a), len(b))
    return sum(a[i] != b[i] for i in range(shortest))

################################################################################
def iflatten(L):
    """Iterative flatten."""
    for sublist in L:
        if hasattr(sublist, '__iter__'):
            for item in iflatten(sublist): yield item
        else: yield sublist

################################################################################
def average(iterator):
    """Iterative mean."""
    count = 0
    total = 0
    for num in iterator:
        count += 1
        total += num
    return float(total)/count

################################################################################
def get_next_item(iterable):
    """Gets the next item of an iterable.
    If the iterable is exhausted, returns None."""
    try: x = iterable.next()
    except StopIteration: x = None
    except AttributeError: x = None
    return x

################################################################################
def pretty_now():
    """Returns some thing like '2014-07-24 11:12:45 CEST+0200'"""
    now = datetime.datetime.now(dateutil.tz.tzlocal())
    return now.strftime("%Y-%m-%d %H:%M:%S %Z%z")

################################################################################
def andify(list_of_strings):
    """
    Given a list of strings will join them with commas
    and a final "and" word.

    >>> andify(['Apples', 'Oranges', 'Mangos'])
    'Apples, Oranges and Mangos'
    """
    result = ', '.join(list_of_strings)
    comma_index = result.rfind(',')
    if comma_index > -1: result = result[:comma_index] + ' and' + result[comma_index+1:]
    return result

################################################################################
def num_to_ith(num):
    """1 becomes 1st, 2 becomes 2nd, etc."""
    value             = str(num)
    before_last_digit = value[-2]
    last_digit        = value[-1]
    if len(value) > 1 and before_last_digit == '1': return value +'th'
    if last_digit == '1': return value + 'st'
    if last_digit == '2': return value + 'nd'
    if last_digit == '3': return value + 'rd'
    return value + 'th'

################################################################################
def isubsample(full_sample, k, full_sample_len=None):
    """Down-sample an enumerable list of things"""
    # Determine length #
    if not full_sample_len: full_sample_len = len(full_sample)
    # Check size coherence #
    if not 0 <= k <= full_sample_len:
        raise ValueError('Required that 0 <= k <= full_sample_length')
    # Do it #
    picked = 0
    for i, element in enumerate(full_sample):
        prob = (k-picked) / (full_sample_len-i)
        if random.random() < prob:
            yield element
            picked += 1
    # Did we pick the right amount #
    assert picked == k

###############################################################################
def moving_average(interval, windowsize, borders=None):
    """This is essentially a convolving operation. Several option exist for dealing with the border cases.

        * None: Here the returned signal will be smaller than the inputted interval.

        * zero_padding: Here the returned signal will be larger than the inputted interval and we will add zeros to the original interval before operating the convolution.

        * zero_padding_and_cut: Same as above only the result is truncated to be the same size as the original input.

        * copy_padding: Here the returned signal will be larger than the inputted interval and we will use the right and leftmost values for padding before operating the convolution.

        * copy_padding_and_cut: Same as above only the result is truncated to be the same size as the original input.

        * zero_stretching: Here we will compute the convolution only in the valid domain, then add zeros to the result so that the output is the same size as the input.

        * copy_stretching: Here we will compute the convolution only in the valid domain, then copy the right and leftmost values so that the output is the same size as the input.
        """
    # The window size in half #
    half = int(math.floor(windowsize/2.0))
    # The normalized rectangular signal #
    window = numpy.ones(int(windowsize))/float(windowsize)
    # How do we deal with borders #
    if borders == None:
        return numpy.convolve(interval, window, 'valid')
    if borders == 'zero_padding':
        return numpy.convolve(interval, window, 'full')
    if borders == 'zero_padding_and_cut':
        return numpy.convolve(interval, window, 'same')
    if borders == 'copy_padding':
        new_interval = [interval[0]]*(windowsize-1) + interval + [interval[-1]]*(windowsize-1)
        return numpy.convolve(new_interval, window, 'valid')
    if borders == 'copy_padding_and_cut':
        new_interval = [interval[0]]*(windowsize-1) + interval + [interval[-1]]*(windowsize-1)
        return numpy.convolve(new_interval, window, 'valid')[half:-half]
    if borders == 'zero_stretching':
        result = numpy.convolve(interval, window, 'valid')
        pad = numpy.zeros(half)
        return numpy.concatenate((pad, result, pad))
    if borders == 'copy_stretching':
        result = numpy.convolve(interval, window, 'valid')
        left = numpy.ones(half)*result[0]
        right = numpy.ones(half)*result[-1]
        return numpy.concatenate((left, result, right))

################################################################################
def wait(predicate, interval=1, message=lambda: "Waiting..."):
    """Wait until the predicate turns true and display a turning ball."""
    ball, next_ball = u"|/-\\", "|"
    sys.stdout.write("    \033[K")
    sys.stdout.flush()
    while not predicate():
        time.sleep(1)
        next_ball = ball[(ball.index(next_ball) + 1) % len(ball)]
        sys.stdout.write("\r " + str(message()) + " " + next_ball + " \033[K")
        sys.stdout.flush()
    print "\r Done. \033[K"
    sys.stdout.flush()

###############################################################################
def natural_sort(item):
    """
    Sort strings that contain numbers correctly.

    >>> l = ['v1.3.12', 'v1.3.3', 'v1.2.5', 'v1.2.15', 'v1.2.3', 'v1.2.1']
    >>> l.sort(key=natural_sort)
    >>> l.__repr__()
    "['v1.2.1', 'v1.2.3', 'v1.2.5', 'v1.2.15', 'v1.3.3', 'v1.3.12']"
    """
    if item is None: return 0
    def try_int(s):
        try: return int(s)
        except ValueError: return s
    return map(try_int, re.findall(r'(\d+|\D+)', item))

###############################################################################
def split_thousands(s, tSep='\'', dSep='.'):
    """
    Splits a number on thousands.
    http://code.activestate.com/recipes/498181-add-thousands-separator-commas-to-formatted-number/

    >>> split_thousands(1000012)
    "1'000'012"
    """
    # Check input #
    if s is None: return 0
    # Check for int #
    if round(s, 13) == s: s = int(s)
    # Make string #
    if not isinstance(s, str): s = str(s)
    # Unreadable code #
    cnt = 0
    numChars = dSep + '0123456789'
    ls = len(s)
    while cnt < ls and s[cnt] not in numChars: cnt += 1
    lhs = s[0:cnt]
    s = s[cnt:]
    if dSep == '': cnt = -1
    else: cnt = s.rfind(dSep)
    if cnt > 0:
        rhs = dSep + s[cnt+1:]
        s = s[:cnt]
    else:
        rhs = ''
    splt=''
    while s != '':
        splt= s[-3:] + tSep + splt
        s = s[:-3]
    return lhs + splt[:-1] + rhs

################################################################################
def gps_deg_to_float(data):
    m = re.search(u"(\d+?)°(\d+?)\'(\d+?)\'\'", data.strip())
    degs, mins, secs = [0.0 if m.group(i) is None else int(m.group(i)) for i in range(1, 4)]
    comp_dir = -1 if data[-1] in ('N', 'E') else 1
    return (degs + (mins / 60) + (secs / 3600)) * comp_dir

################################################################################
def is_integer(string):
    try: int(string)
    except ValueError: return False
    return True

################################################################################
def reverse_compl_with_name(old_seq):
    """Reverse a SeqIO sequence, but keep its name intact."""
    new_seq = old_seq.reverse_complement()
    new_seq.id = old_seq.id
    new_seq.description = old_seq.description
    return new_seq

################################################################################
class GenWithLength(object):
    """A generator with a length attribute"""
    def __init__(self, gen, length): self.gen, self.length = gen, length
    def __iter__(self): return self.gen
    def __len__(self):  return self.length

###############################################################################
class Password(object):
    """A password object that will only prompt the user once per session"""
    def __str__(self): return self.value
    def __init__(self, prompt=None):
        self._value = None
        self.prompt = prompt

    @property
    def value(self):
        if self._value == None: self._value = getpass.getpass(self.prompt)
        return self._value

################################################################################
class OrderedSet(collections.OrderedDict, collections.MutableSet):
    """A recipe for an ordered set.
    http://stackoverflow.com/a/1653978/287297"""

    def update(self, *args, **kwargs):
        if kwargs:
            raise TypeError("update() takes no keyword arguments")

        for s in args:
            for e in s:
                self.add(e)

    def add(self, elem):
        self[elem] = None

    def discard(self, elem):
        self.pop(elem, None)

    def __le__(self, other):
        return all(e in other for e in self)

    def __lt__(self, other):
        return self <= other and self != other

    def __ge__(self, other):
        return all(e in self for e in other)

    def __gt__(self, other):
        return self >= other and self != other

    def __repr__(self):
        return 'OrderedSet([%s])' % (', '.join(map(repr, self.keys())))

    def __str__(self):
        return '{%s}' % (', '.join(map(repr, self.keys())))

    difference = property(lambda self: self.__sub__)
    difference_update = property(lambda self: self.__isub__)
    intersection = property(lambda self: self.__and__)
    intersection_update = property(lambda self: self.__iand__)
    issubset = property(lambda self: self.__le__)
    issuperset = property(lambda self: self.__ge__)
    symmetric_difference = property(lambda self: self.__xor__)
    symmetric_difference_update = property(lambda self: self.__ixor__)
    union = property(lambda self: self.__or__)

################################################################################
class SuppressAllOutput(object):
    """For those annoying modules that can't shut-up about warnings."""

    def __enter__(self):
        # Standard error #
        sys.stderr.flush()
        self.old_stderr = sys.stderr
        sys.stderr = open('/dev/null', 'a+', 0)
        # Standard out #
        sys.stdout.flush()
        self.old_stdout = sys.stdout
        sys.stdout = open('/dev/null', 'a+', 0)

    def __exit__(self, exc_type, exc_value, traceback):
        # Standard error #
        sys.stderr.flush()
        sys.stderr = self.old_stderr
        # Standard out #
        sys.stdout.flush()
        sys.stdout = self.old_stdout

    def test():
        print >>sys.stdout, "printing to stdout before suppression"
        print >>sys.stderr, "printing to stderr before suppression"
        with SuppressAllOutput():
            print >>sys.stdout, "printing to stdout during suppression"
            print >>sys.stderr, "printing to stderr during suppression"
        print >>sys.stdout, "printing to stdout after suppression"
        print >>sys.stderr, "printing to stderr after suppression"

################################################################################
def load_json_path(path):
    """Load a file with the json module, but report errors better if it
    fails. And have it ordered too !"""
    with open(path) as handle:
        try: return json.load(handle, object_pairs_hook=collections.OrderedDict)
        except ValueError as error:
            message = "Could not decode JSON file '%s'." % path
            message = "-"*20 + "\n" + message + "\n" + str(error) + "\n" + "-"*20 + "\n"
            sys.stderr.write(message)
            raise error

###############################################################################
def find_file_by_name(name, root=os.curdir):
    for dirpath, dirnames, filenames in os.walk(os.path.abspath(root)):
        if name in filenames: return os.path.join(dirpath, name)
    raise Exception("Could not find file '%s' in '%s'" % (name, root))

################################################################################
def md5sum(file_path, blocksize=65536):
    """Compute the md5 of a file. Pretty fast."""
    md5 = hashlib.md5()
    with open(file_path, "rb") as f:
        for block in iter(lambda: f.read(blocksize), ""):
            md5.update(block)
    return md5.hexdigest()

################################################################################
def download_from_url(source, destination, progress=False, uncompress=True):
    """Download a file from an URL and place it somewhere. Like wget.
    Uses requests and tqdm to display progress if you want.
    By default it will uncompress files."""
    from tqdm import tqdm
    import requests
    response = requests.get(source, stream=True)
    with open(destination, "wb") as handle:
        if progress:
            for data in tqdm(response.iter_content()): handle.write(data)
        else:
            for data in response.iter_content(): handle.write(data)
    if uncompress:
        with open(destination, 'r') as f: header = f.read(4)
        if header == "PK\x03\x04": unzip(destination, inplace=True)
        # Add other compression formats here

################################################################################
def unzip(source, destination=None, inplace=False, single=True):
    """Unzip a standard zip file. Can specify the destination of the
    uncompressed file, or just set inplace=True to delete the original."""
    # Load #
    import zipfile, tempfile, shutil
    # Check #
    assert zipfile.is_zipfile(source)
    # Load #
    z = zipfile.ZipFile(source)
    if single or inplace: assert len(z.infolist()) == 1
    # Single file #
    if single:
        member = z.infolist()[0]
        tmpdir = tempfile.mkdtemp() + '/'
        z.extract(member, tmpdir)
        z.close()
        if inplace: shutil.move(tmpdir + member.filename, source)
        else:       shutil.move(tmpdir + member.filename, destination)
    # Multifile - no security, dangerous #
    if not single:
        z.extractall()

###############################################################################
def reversed_lines(path):
    """Generate the lines of file in reverse order."""
    with open(path, 'r') as handle:
        part = ''
        for block in reversed_blocks(handle):
            for c in reversed(block):
                if c == '\n' and part:
                    yield part[::-1]
                    part = ''
                part += c
        if part: yield part[::-1]

###############################################################################
def reversed_blocks(handle, blocksize=4096):
    """Generate blocks of file's contents in reverse order."""
    handle.seek(0, os.SEEK_END)
    here = handle.tell()
    while 0 < here:
        delta = min(blocksize, here)
        here -= delta
        handle.seek(here, os.SEEK_SET)
        yield handle.read(delta)

################################################################################
def prepend_to_file(path, data, bufsize=1<<15):
    # Backup the file #
    backupname = path + os.extsep + 'bak'
    # Remove previous backup if it exists #
    try: os.unlink(backupname)
    except OSError: pass
    os.rename(path, backupname)
    # Open input/output files,  note: outputfile's permissions lost #
    with open(backupname) as inputfile:
        with open(path, 'w') as outputfile:
            outputfile.write(data)
            buf = inputfile.read(bufsize)
            while buf:
                outputfile.write(buf)
                buf = inputfile.read(bufsize)
    # Remove backup on success #
    os.remove(backupname)

def append_to_file(path, data):
    with open(path, "a") as handle:
        handle.write(data)

################################################################################
def tail(path, window=20):
    with open(path, 'r') as f:
        BUFSIZ = 1024
        f.seek(0, 2)
        num_bytes = f.tell()
        size = window + 1
        block = -1
        data = []
        while size > 0 and num_bytes > 0:
            if num_bytes - BUFSIZ > 0:
                # Seek back one whole BUFSIZ
                f.seek(block * BUFSIZ, 2)
                # Read BUFFER
                data.insert(0, f.read(BUFSIZ))
            else:
                # File too small, start from beginning
                f.seek(0,0)
                # Only read what was not read
                data.insert(0, f.read(num_bytes))
            linesFound = data[0].count('\n')
            size -= linesFound
            num_bytes -= BUFSIZ
            block -= 1
        return '\n'.join(''.join(data).splitlines()[-window:])

def head(path, lines=20):
    with open(path, 'r') as handle:
        return ''.join(handle.next() for line in xrange(lines))

###############################################################################
def which(cmd, safe=False):
    """https://github.com/jc0n/python-which"""
    from plumbing.autopaths import FilePath
    def is_executable(path):
        return os.path.exists(path) and os.access(path, os.X_OK) and not os.path.isdir(path)
    path, name = os.path.split(cmd)
    if path:
        if is_executable(cmd): return FilePath(cmd)
    else:
        for path in os.environ['PATH'].split(os.pathsep):
            candidate = os.path.join(path, cmd)
            if is_executable(candidate): return FilePath(candidate)
    if not safe: raise Exception('which failed to locate a proper command path "%s"' % cmd)



# Built-in modules #
import tempfile

# Internal modules #
from plumbing.autopaths import FilePath, DirectoryPath

################################################################################
def new_temp_path(**kwargs):
    """A new temporary path"""
    handle = tempfile.NamedTemporaryFile(**kwargs)
    path = handle.name
    handle.close()
    return path

################################################################################
def new_temp_file(**kwargs):
    """A new temporary file"""
    return TmpFile.empty(**kwargs)

################################################################################
def new_temp_dir(**kwargs):
    return DirectoryPath(tempfile.mkdtemp() + '/')

################################################################################
class TmpFile(FilePath):
    def __repr__(self): return self.path

    @classmethod
    def empty(cls, **kwargs): return cls(**kwargs)

    @classmethod
    def from_string(cls, string, **kwargs): return cls(content=string, **kwargs)

    def __enter__(self):
        self.handle = open(self.path, 'w')
        return self

    def __exit__(self, *exc_info):
        self.handle.close()

    def __new__(cls, path=None, content=None, **kwargs):
        handle = open(path, 'w') if path else tempfile.NamedTemporaryFile(delete=False, **kwargs)
        if content: handle.write(content)
        handle.close()
        return FilePath.__new__(cls, handle.name)

    def __init__(self, path=None, content=None, **kwargs):
        if not path: path = str(self)
        self.path = path


# Built-in modules #
import os, stat, tempfile, re, subprocess, shutil, codecs, gzip
import glob, warnings

# Internal modules #
from plumbing.common import append_to_file, prepend_to_file
from plumbing.common import md5sum, natural_sort, unzip

# Third party modules #
import sh

################################################################################
class AutoPaths(object):
    """
    You can use this class like this when making pipelines:

        class Sample(object):
            all_paths = '''
                /raw/raw.sff
                /raw/raw.fastq
                /clean/trim.fastq
                /clean/clean.fastq'''

            def __init__(self, base_dir):
                self.p = AutoPaths(base_dir, self.all_paths)

            def clean(self):
                shutil.move(self.p.raw_sff, self.p.clean_fastq)
    """

    def __repr__(self): return '<%s object on "%s">' % (self.__class__.__name__, self._base_dir)

    def __init__(self, base_dir, all_paths):
        # Don't nest DirectoryPaths or the like #
        if hasattr(base_dir, 'path'): base_dir = base_dir.path
        # Attributes #
        self._base_dir  = base_dir
        self._all_paths = all_paths
        self._tmp_dir   = tempfile.gettempdir() + '/'
        # Parse input #
        self._paths = [PathItems(p.lstrip(' '), base_dir) for p in all_paths.split('\n')]

    def __call__(self, key):    return self.__getattr__(key)
    def __getitem__(self, key): return self.__getattr__(key)

    def __getattr__(self, key):
        # Let built-ins pass through to object #
        if key.startswith('__') and key.endswith('__'): return object.__getattr__(key)
        # Special cases that should do to the actual dictionary #
        if key.startswith('_'): return self.__dict__[key]
        # Temporary items #
        if key == 'tmp_dir': return self.__dict__['_tmp_dir']
        if key == 'tmp': return self.__dict__['tmp']
        # Search #
        items = key.split('_')
        # Is it a directory ? #
        if items[-1] == 'dir':
            items.pop(-1)
            return self.search_for_dir(key, items)
        else:
            return self.search_for_file(key, items)

    def search_for_file(self, key, items):
        # Search #
        matches = [set([p for p in self._paths if i in p]) for i in items]
        result = set.intersection(*matches)
        # No matches #
        if len(result) == 0:
            raise Exception("Could not find any path matching '%s'" % key)
        # Multiple matches, advantage file name #
        if len(result) > 1:
            best_score = max([p.score_file(items) for p in result])
            result = [p for p in result if p.score_file(items) >= best_score]
        # Multiple matches, take the one with less parts #
        if len(result) > 1:
            shortest = min([len(p) for p in result])
            result = [p for p in result if len(p) <= shortest]
        # Multiple matches, error #
        if len(result) > 1:
            raise Exception("Found several paths matching '%s'" % key)
        # Make the directory #
        result = result.pop()
        directory = DirectoryPath(result.complete_dir)
        if not directory.exists: directory.create(safe=True)
        # Return #
        return FilePath(result.complete_path)

    def search_for_dir(self, key, items):
        # Search #
        matches = [set([p for p in self._paths if i in p]) for i in items]
        result = set.intersection(*matches)
        # No matches #
        if len(result) == 0:
            raise Exception("Could not find any path matching '%s'" % key)
        # Multiple matches, advantage dir name #
        if len(result) > 1:
            best_score = max([p.score_dir(items) for p in result])
            result = [p for p in result if p.score_dir(items) >= best_score]
        # Multiple matches, take the one with less parts #
        if len(result) > 1:
            shortest = min([len(p) for p in result])
            result = [p for p in result if len(p) <= shortest]
        # Multiple matches, maybe they all are the same directory #
        if len(result) > 1:
            if len(set([p.dir for p in result])) == 1: result = [result[0]]
        # Multiple matches, error #
        if len(result) > 1:
            raise Exception("Found several paths matching '%s'" % key)
        # Make the directory #
        result = result.pop()
        directory = DirectoryPath(result.complete_dir)
        if not directory.exists: directory.create(safe=True)
        # Return #
        return directory

    @property
    def tmp_dir(self):
        if not self._tmp_dir: self._tmp_dir = tempfile.mkdtemp() + '/'
        return self._tmp_dir

    @property
    def tmp(self):
        return self.tmp_dir + 'autopath.tmp'

################################################################################
class PathItems(object):
    delimiters = '_', '.', '/'
    pattern = re.compile('|'.join(map(re.escape, delimiters)))

    def __repr__(self): return '<%s object "%s">' % (self.__class__.__name__, self.path)

    def __init__(self, path, base_dir):
        self.path = path
        self.base_dir = base_dir
        self.dir, self.name = os.path.split(path)
        self.name_items = self.pattern.split(self.name) if self.name else []
        self.dir_items = self.pattern.split(self.dir) if self.dir else []
        self.all_items = self.name_items + self.dir_items

    def __contains__(self, i):
        return i in self.all_items

    def __len__(self):
        return len(self.all_items)

    def score_file(self, items):
        return sum([1.0 if i in self.name_items else 0.5 for i in items if i in self])

    def score_dir(self, items):
        return sum([1.0 if i in self.dir_items else 0.5 for i in items if i in self])

    @property
    def complete_path(self):
        return '/' + os.path.relpath(self.base_dir + self.path, '/')

    @property
    def complete_dir(self):
        return '/' + os.path.relpath(self.base_dir + self.dir, '/') + '/'

################################################################################
class DirectoryPath(str):

    def __repr__(self): return '<%s object "%s">' % (self.__class__.__name__, self.path)

    @classmethod
    def clean_path(cls, path):
        """Given a path, return a cleaned up version for initialization."""
        # Conserve 'None' object style #
        if path is None: return None
        # Don't nest DirectoryPaths or the like #
        if hasattr(path, 'path'): path = path.path
        # Expand the tilda #
        if "~" in path: path = os.path.expanduser(path)
        # Our standard is to end with a slash for directories #
        if not path.endswith('/'): path += '/'
        # Return the result #
        return path

    def __new__(cls, path, *args, **kwargs):
        """A DirectoryPath is in fact a string"""
        return str.__new__(cls, cls.clean_path(path))

    def __init__(self, path):
        self.path = self.clean_path(path)

    def __add__(self, other):
        return self.path + other

    @property
    def p(self):
        if not hasattr(self, 'all_paths'):
            raise Exception("You need to define 'all_paths' to use this function")
        return AutoPaths(self.path, self.all_paths)

    @property
    def name(self):
        """Just the directory name"""
        return os.path.basename(os.path.dirname(self.path))

    @property
    def prefix_path(self):
        """The full path without the extension"""
        return os.path.splitext(self.path)[0].rstrip('/')

    @property
    def absolute_path(self):
        """The absolute path starting with a `/`"""
        return os.path.abspath(self.path) + '/'

    @property
    def directory(self):
        """The full path of the directory containing this one."""
        return DirectoryPath(os.path.dirname(os.path.dirname(self.path)))

    #-------------------------- Recursive contents ---------------------------#
    @property
    def contents(self):
        """The files and directories in this directory, recursively."""
        for root, dirs, files in os.walk(self.path, topdown=False):
            for d in dirs:  yield DirectoryPath(os.path.join(root, d))
            for f in files: yield FilePath(os.path.join(root, f))

    @property
    def files(self):
        """The files in this directory, recursively."""
        for root, dirs, files in os.walk(self.path, topdown=False):
            for f in files: yield FilePath(os.path.join(root, f))

    @property
    def directories(self):
        """The directories in this directory, recursively."""
        for root, dirs, files in os.walk(self.path, topdown=False):
            for d in dirs: yield DirectoryPath(os.path.join(root, d))

    #----------------------------- Flat contents -----------------------------#
    @property
    def flat_contents(self):
        """The files and directories in this directory non-recursively."""
        for root, dirs, files in os.walk(self.path):
            for d in dirs:  yield DirectoryPath(os.path.join(root, d))
            for f in files: yield FilePath(os.path.join(root, f))
            break

    @property
    def flat_files(self):
        """The files in this directory non-recursively, and sorted.
        #TODO: check for permission denied in directory."""
        result = []
        for root, dirs, files in os.walk(self.path):
            result = [FilePath(os.path.join(root, f)) for f in files]
            break
        result.sort(key=lambda x: natural_sort(x.path))
        return result

    @property
    def flat_directories(self):
        """The directories in this directory non-recursively, and sorted."""
        result = []
        for root, dirs, files in os.walk(self.path):
            result = [DirectoryPath(os.path.join(root, d)) for d in dirs]
            break
        result.sort(key=lambda x: natural_sort(x.path))
        return result

    #-------------------------------- Other ----------------------------------#
    @property
    def is_symlink(self):
        """Is this directory a symbolic link to an other directory?"""
        return os.path.islink(self.path.rstrip('/'))

    @property
    def exists(self):
        """Does it exist in the file system?"""
        return os.path.lexists(self.path) # Include broken symlinks

    @property
    def empty(self):
        """Does the directory contain no files?"""
        return len(list(self.flat_contents)) == 0

    @property
    def permissions(self):
        """Convenience object for dealing with permissions."""
        return FilePermissions(self.path)

    @property
    def mod_time(self):
        """The modification time"""
        return os.stat(self.path).st_mtime

    def remove(self):
        if not self.exists: return False
        if self.is_symlink: return self.remove_when_symlink()
        shutil.rmtree(self.path, ignore_errors=True)
        return True

    def remove_when_symlink(self):
        if not self.exists: return False
        os.remove(self.path.rstrip('/'))
        return True

    def create(self, safe=False, inherit=True):
        # Create it #
        if not safe:
            os.makedirs(self.path)
            if inherit: os.chmod(self.path, self.directory.permissions.number)
        if safe:
            try:
                os.makedirs(self.path)
                if inherit: os.chmod(self.path, self.directory.permissions.number)
            except OSError: pass

    def create_if_not_exists(self):
        if not self.exists: self.create()

    def zip(self, keep_orig=False):
        """Make a zip archive of the directory"""
        shutil.make_archive(self.prefix_path , "zip", self.directory, self.name)
        if not keep_orig: self.remove()

    def link_from(self, where, safe=False):
        """Make a link here pointing to another directory somewhere else.
        The destination is hence self.path and the source is *where*"""
        if not safe:
            self.remove()
            return os.symlink(where, self.path.rstrip('/'))
        if safe:
            try: self.remove()
            except OSError: pass
            try: os.symlink(where, self.path.rstrip('/'))
            except OSError: warnings.warn("Symlink of %s to %s did not work" % (where, self))

    def copy(self, path):
        assert not os.path.exists(path)
        shutil.copytree(self.path, path)

    def glob(self, pattern):
        """Perform a glob search in this directory."""
        files = glob.glob(self.path + pattern)
        return map(FilePath, files)

    def find(self, pattern):
        """Find a file in this directory."""
        f = glob.glob(self.path + pattern)[0]
        return FilePath(f)

################################################################################
class FilePath(str):
    """I can never remember all those darn `os.path` commands, so I made a class
    that wraps them with an easier and more pythonic syntax.

        path = FilePath('/home/root/text.txt')
        print path.extension
        print path.directory
        print path.filename

    You can find lots of the common things you would need to do with file paths.
    Such as: path.make_executable() etc etc."""

    def __repr__(self):    return '<%s object "%s">' % (self.__class__.__name__, self.path)
    def __nonzero__(self): return self.path != None and self.count_bytes != 0
    def __list__(self):    return self.count
    def __iter__(self):
        with open(self.path, 'r') as handle:
            for line in handle: yield line
    def __len__(self):
        if self.path is None: return 0
        return self.count

    def __new__(cls, path, *args, **kwargs):
        """A FilePath is in fact a string"""
        return str.__new__(cls, cls.clean_path(path))

    def __init__(self, path):
        self.path = self.clean_path(path)

    def __sub__(self, directory):
        """Subtract a directory from the current path to get the relative path
        of the current file from that directory."""
        return os.path.relpath(self.path, directory)

    @classmethod
    def clean_path(cls, path):
        """Given a path, return a cleaned up version for initialization"""
        # Conserve None object style #
        if path is None: return None
        # Don't nest FilePaths or the like #
        if hasattr(path, 'path'): path = path.path
        # Expand tilda #
        if "~" in path: path = os.path.expanduser(path)
        # Expand star #
        if "*" in path:
            matches = glob.glob(path)
            if len(matches) < 1: raise Exception("Found exactly no files matching '%s'" % path)
            if len(matches) > 1: raise Exception("Found several files matching '%s'" % path)
            path = matches[0]
        # Return the result #
        return path

    @property
    def first(self):
        """Just the first line"""
        with open(self.path, 'r') as handle:
            for line in handle: return line

    @property
    def exists(self):
        """Does it exist in the file system"""
        return os.path.lexists(self.path) # Returns True even for broken symbolic links

    @property
    def prefix_path(self):
        """The full path without the (last) extension and trailing period"""
        return str(os.path.splitext(self.path)[0])

    @property
    def prefix(self):
        """Just the filename without the (last) extension and trailing period"""
        return str(os.path.basename(self.prefix_path))

    @property
    def short_prefix(self):
        """Just the filename without any extension or periods"""
        return self.filename.split('.')[0]

    @property
    def filename(self):
        """Just the filename with the extension"""
        return str(os.path.basename(self.path))

    @property
    def directory(self):
        """The directory containing this file"""
        # The built-in function #
        directory = os.path.dirname(self.path)
        # Maybe we need to go the absolute path way #
        if not directory: directory = os.path.dirname(self.absolute_path)
        # Return #
        return DirectoryPath(directory + '/')

    @property
    def extension(self):
        """The extension with the leading period"""
        return os.path.splitext(self.path)[1]

    @property
    def count_bytes(self):
        """The number of bytes"""
        if not self.exists: return 0
        return os.path.getsize(self.path)

    @property
    def count(self):
        """We are going to default to the number of lines"""
        return int(sh.wc('-l', self.path).split()[0])

    @property
    def size(self):
        """Human readable file size"""
        return Filesize(self.count_bytes)

    @property
    def permissions(self):
        """Convenience object for dealing with permissions"""
        return FilePermissions(self.path)

    @property
    def contents(self):
        """The contents as a string"""
        with open(self.path, 'r') as handle: return handle.read()

    @property
    def absolute_path(self):
        """The absolute path starting with a `/`"""
        return FilePath(os.path.abspath(self.path))

    @property
    def physical_path(self):
        """The physical path like in `pwd -P`"""
        return FilePath(os.path.realpath(self.path))

    @property
    def relative_path(self):
        """The relative path when compared with current directory"""
        return FilePath(os.path.relpath(self.physical_path))

    @property
    def md5(self):
        """Return the md5 checksum."""
        return md5sum(self.path)

    @property
    def might_be_binary(self):
        """Try to quickly guess if the file is binary."""
        from binaryornot.check import is_binary
        return is_binary(self.path)

    @property
    def contains_binary(self):
        """Return True if the file contains binary characters."""
        from binaryornot.helpers import is_binary_string
        return is_binary_string(self.contents)

    def read(self, encoding=None):
        with codecs.open(self.path, 'r', encoding) as handle: content = handle.read()
        return content

    def create(self):
        with open(self.path, 'w'): pass

    def write(self, content, encoding=None):
        if encoding is None:
            with open(self.path, 'w') as handle: handle.write(content)
        else:
            with codecs.open(self.path, 'w', encoding) as handle: handle.write(content)

    def writelines(self, content, encoding=None):
        if encoding is None:
            with open(self.path, 'w') as handle: handle.writelines(content)
        else:
            with codecs.open(self.path, 'w', encoding) as handle: handle.writelines(content)

    def remove(self):
        if not self.exists: return False
        os.remove(self.path)
        return True

    def copy(self, path):
        shutil.copy2(self.path, path)

    def execute(self):
        return subprocess.call([self.path])

    def replace_extension(self, new_extension='txt'):
        """Return a new path with the extension swapped out"""
        return FilePath(os.path.splitext(self.path)[0] + '.' + new_extension)

    def new_name_insert(self, string):
        """Return a new name by appending a string before the extension"""
        return self.prefix_path + "." + string + self.extension

    def make_directory(self):
        """Create the directory the file is supposed to be in if it does not exist"""
        if not self.directory.exists: self.directory.create()

    def must_exist(self):
        """Raise an exception if the path doesn't exist."""
        if not self.exists: raise Exception("The file path '%s' does not exist." % self.path)

    def head(self, lines=10):
        """Return the first few lines."""
        content = iter(self)
        for x in xrange(lines):
            yield content.next()

    def move_to(self, path):
        """Move the file."""
        assert not os.path.exists(path)
        shutil.move(self.path, path)

    def link_from(self, path, safe=False):
        """Make a link here pointing to another file somewhere else.
        The destination is hence self.path and the source is *path*."""
        # Get source and destination #
        source      = path
        destination = self.path
        # Do it #
        if not safe:
            if os.path.exists(destination): os.remove(destination)
            os.symlink(source, destination)
        # Do it safely #
        if safe:
            try: os.remove(destination)
            except OSError: pass
            try: os.symlink(source, destination)
            except OSError: pass

    def link_to(self, path, safe=False, absolute=True):
        """Create a link somewhere else pointing to this file.
        The destination is hence *path* and the source is self.path."""
        # Get source and destination #
        if absolute: source = self.absolute_path
        else:        source = self.path
        destination = path
        # Do it #
        if not safe:
            if os.path.exists(destination): os.remove(destination)
            os.symlink(source, destination)
        # Do it safely #
        if safe:
            try: os.remove(destination)
            except OSError: pass
            try: os.symlink(source, destination)
            except OSError: pass

    def gzip_to(self, path=None):
        """Make a gzipped version of the file at a given path"""
        if path is None: path = self.path + ".gz"
        with open(self.path, 'rb') as orig_file:
            with gzip.open(path, 'wb') as new_file:
                new_file.writelines(orig_file)
        return FilePath(path)

    def ungzip_to(self, path=None):
        """Make an unzipped version of the file at a given path"""
        if path is None: path = self.path[:3]
        with gzip.open(self, 'rb') as orig_file:
            with open(path, 'wb') as new_file:
                new_file.writelines(orig_file)
        return FilePath(path)

    def zip_to(self, path=None):
        """Make a zipped version of the file at a given path."""
        pass

    def unzip_to(self, destination=None, inplace=False):
        """Make an unzipped version of the file at a given path"""
        return unzip(self.path, destination=destination, inplace=inplace)

    def append(self, what):
        """Append some text or an other file to the current file"""
        if isinstance(what, FilePath): what = what.contents
        append_to_file(self.path, what)

    def prepend(self, what):
        """Append some text or an other file to the current file"""
        if isinstance(what, FilePath): what = what.contents
        prepend_to_file(self.path, what)

################################################################################
class Filesize(object):
    """
    Container for a size in bytes with a human readable representation
    Use it like this:

        >>> size = Filesize(123123123)
        >>> print size
        '117.4 MiB'
    """

    chunk      = 1000 # Could be 1024 if you like old-style sizes
    units      = ['bytes', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB']
    precisions = [0, 0, 1, 2, 2, 2]

    def __init__(self, size):
        self.size = size

    def __int__(self):
        return self.size

    def __eq__(self, other):
        return self.size == other

    def __str__(self):
        if self.size == 0: return '0 bytes'
        from math import log
        unit = self.units[min(int(log(self.size, self.chunk)), len(self.units) - 1)]
        return self.format(unit)

    def format(self, unit):
        # Input checking #
        if unit not in self.units: raise Exception("Not a valid file size unit: %s" % unit)
        # Special no plural case #
        if self.size == 1 and unit == 'bytes': return '1 byte'
        # Compute #
        exponent      = self.units.index(unit)
        quotient      = float(self.size) / self.chunk**exponent
        precision     = self.precisions[exponent]
        format_string = '{:.%sf} {}' % (precision)
        # Return a string #
        return format_string.format(quotient, unit)

################################################################################
class FilePermissions(object):
    """Container for reading and setting a files permissions"""

    def __init__(self, path):
        self.path = path

    @property
    def number(self):
        """The permission bits as an octal integer"""
        return os.stat(self.path).st_mode & 0777

    def make_executable(self):
        return os.chmod(self.path, os.stat(self.path).st_mode | stat.S_IEXEC)

    def only_readable(self):
        """Remove all writing privileges"""
        return os.chmod(self.path, stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)



# Built-in modules #
import os, sys

# Internal modules #
from plumbing.autopaths import DirectoryPath

# Third party modules #
import sh

###############################################################################
class GitRepo(DirectoryPath):
    """A git repository with some convenience methods.
    Requires at least git 2.7 (release January 5th, 2015)."""

    def __nonzero__(self):
        return os.path.exists(self.git_dir)

    def __init__(self, path, empty=False):
        # Super #
        DirectoryPath.__init__(self, path)
        # The git directory #
        self.git_dir = self.path + '.git'
        # Check exists #
        if not empty and not self:
            raise Exception("No git repository at '%s'" % (self.git_dir))
        # Default arguments #
        self.default = ["--git-dir=" + self.git_dir, "--work-tree=" + self.path]

    #------------------------------- Properties ------------------------------#
    @property
    def tag(self):
        """For instance: u'1.0.3-69-gf0c796d-dirty'"""
        tag = sh.git(self.default + ["describe", "--tags", "--dirty", "--always"])
        return tag.strip('\n')

    @property
    def hash(self):
        """For instance: u'f0c796dae64a5a118d88e60523c011d535e8c476'"""
        sha1 = sh.git(self.default + ["rev-parse", "HEAD"])
        return sha1.strip('\n')

    @property
    def short_hash(self):
        """For instance: u'f0c796d'"""
        sha1 = sh.git(self.default + ["rev-parse", "--short", "HEAD"])
        return sha1.strip('\n')

    @property
    def branch(self):
        """For instance: u'master'"""
        result = sh.git(self.default + ['symbolic-ref', '--short', 'HEAD'])
        return result.strip('\n')

    @property
    def branches(self):
        """All branches in a list"""
        result = sh.git(self.default + ['branch', '-a', '--no-color'])
        return [l.strip(' *\n') for l in result.split('\n') if l.strip(' *\n')]

    @property
    def remote_branch(self):
        """For instance: u'origin/master'"""
        result = sh.git(self.default + ['rev-parse', '--symbolic-full-name', '--abbrev-ref', '@{u}'])
        return result.strip('\n')

    @property
    def remote_url(self):
        """For instance: u'origin/master'"""
        result = sh.git(self.default + ['remote', 'get-url', 'origin'])
        return result.strip('\n')

    #--------------------------------- Methods -------------------------------#
    def re_clone(self, repo_dir):
        """Clone again, somewhere else"""
        sh.git('clone', self.remote_url, repo_dir)
        return GitRepo(repo_dir)

    def add(self, what):
        return sh.git(self.default + ['add', what])

    def commit(self, message):
        return sh.git(self.default + ['commit', '-m', '"' + message + '"'])

    def push(self, source=None, destination=None, tags=False, shell=False):
        # Command #
        command = self.default + ['push']
        # Tags #
        if tags: command.append('--tags')
        # Source and dest #
        if source:                 command.append(source)
        if source and destination: command.append(destination)
        # Show on shell #
        if shell: return sh.git(command, _out=sys.stdout, _err=sys.stderr)
        else:     return sh.git(command)

    def tag_head(self, tag):
        return sh.git(self.default + ['tag', tag, 'HEAD'])


"""
================
Warping programs
================

The ``@command`` decorator makes binding command line executables
easy to write and easy to use.

To wrap a program, write a function that takes whatever arguments
you will need to run the program like if it where on the shell.
The function should return a dictionary containing several keys:

    * mandatory: ``arguments``
    * optional: ``stdin``
    * optional: ``return_value``.

Firstly, ``arguments`` should be a list of strings which is
the actual command and arguments to be executed (e.g. ``["touch", filename]``).

Secondly, ``stdin`` should be a value to feed the subprocess once it is launched.

Thirdly, ``return_value`` should be a value to return, or a callable
object which takes ``(stdout,strerr)`` as parameters and returns the value
that will be passed back to the user when this program is run.
You can also simply specify ``"stdout"`` to have the output of the
process returned directly.

For example, to wrap ``touch``, we write a one argument function that
takes the filename of the file to touch, and apply the ``@command``
decorator to it::

    from plumbing.common import command

    @command
    def touch(filename):
        return {"arguments": ["touch", filename],
                "return_value": filename}

We can now call this function directly::

     f = touch("myfile")

The value returned by touch is ``"myfile"``, the name of
the touched file.

A more complicated example would include binding the BLASTP algorithm::

    from plumbing.common import command

    @command
    def blastp(database, sequences, **kwargs):
         \"\"\"Will launch the 'blastp' algorithm using the NCBI executable.

        :param database: The path to the database to blast against.
        :param sequences: A fasta formated string.
        :param **kwargs: Extra parameters that will be passed to the executable
                        For instance, you could specify "e=1e-20".
        :returns: A list of top hits in blast format.
        \"\"\"
        return {"arguments": ["blastall", "-p", "blastp", "-d" database] +
                             [a for k,v in kwargs.items() for a in ('-'+k,v)],
                "stdin": sequences)
                "return_value": 'stdout'}

As shown above, we can now call this function directly::

     hits = blastp("swissprot", open("proteins.fasta").read())

The value returned by blastp is a long string containing all the
results from the BLASTP algorithm.

Often you want to call a function, but not block when it returns
so you can run several in parallel. ``@command`` also creates a
method ``parallel`` which does this. The return value is a
Future object with a single method: ``wait()``. When you call
``wait()``, it blocks until the program finishes, then returns the
same value that you would get from calling the function directly.
So, to touch two files, and not block until both commands have
started, you would write::

    a = blastp.parallel("nr", open("fasta1").read(), e=1e-20)
    b = blastp.parallel("nr", open("fasta2").read(), e=1e-20)
    hitsA = a.wait()
    hitsB = b.wait()

The ``parallel`` method will runs processes without blocking.
Other methods exists for running commands in parallel.
For example, on systems using the SLURM batch submission
system, you can run commands via batch submission by using the
``slurm`` method and optionally adding the time and account info::

    p = blastp.slurm("nr", open("fasta").read(), time='1:00:00')
    hits = p.wait()

On systems using the LSF batch submission system, you can run
commands via batch submission by using the ``lsf`` method::

    p = blastp.lsf("nr", open("fasta").read(), queue='long')
    hits = p.wait()
"""

# Built-in modules #
import subprocess, sys, time

# Variables #
PARRALEL_JOBS = []

################################################################################
def start_process(args):
    """Run a process using subprocess module"""
    try:
        return subprocess.Popen(args, bufsize=-1, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE)
    except OSError:
        raise ValueError("Program '%s' does not seem to exist in your $PATH." % args[0])

################################################################################
def pause_for_parallel_jobs(update_interval=2):
    """Wait until all parallel jobs are done and print a status update"""
    global PARRALEL_JOBS
    while True:
        PARRALEL_JOBS = [job for job in PARRALEL_JOBS if not job.finished]
        if not PARRALEL_JOBS:
            sys.stdout.write("\r\033[K")
            sys.stdout.flush()
            return
        sys.stdout.write("\r    %i parallel jobs still running.\033[K" % len(PARRALEL_JOBS))
        sys.stdout.flush()
        time.sleep(update_interval)

################################################################################
class CommandFailed(Exception):
    """Thrown when a program bound by ``@command``
    exits with a value other than ``0``."""

    def __init__(self, args, stderr, name=None):
        message = "Running '%s' failed." % " ".join(args)
        if name: message += " The job name was: '%s'." % name
        if stderr: message += " The error reported is:\n\n" + stderr
        Exception.__init__(self, message)

################################################################################
class command(object):
    """Decorator used to wrap external programs."""

    def __init__(self, function):
        self.function = function
        self.__doc__ = function.__doc__
        self.__name__ = function.__name__

    def __call__(self, *args, **kwargs):
        """Run a program locally, and block until it completes."""
        # Call the user defined function #
        cmd_dict = self.function(*args, **kwargs)
        cmd_dict['arguments'] = [str(a) for a in cmd_dict['arguments']]
        args = cmd_dict['arguments']
        # Start a process #
        proc = start_process(args)
        # Wait until completion #
        try: stdout, stderr = proc.communicate(cmd_dict.get("stdin"))
        except KeyboardInterrupt as err:
            print "You aborted the process pid %i. It was: %s " % (proc.pid, args)
            raise err
        # Check for failure #
        if proc.returncode != 0: raise CommandFailed(args, stderr)
        # Return result #
        result = cmd_dict.get("return_value")
        if callable(result): return result(stdout, stderr)
        elif result == 'stdout': return stdout
        else: return result

    def parallel(self, *args, **kwargs):
        """Run a program and return a Future object."""
        # Call the user defined function #
        cmd_dict = self.function(*args, **kwargs)
        cmd_dict['arguments'] = [str(a) for a in cmd_dict['arguments']]
        # Start a process #
        proc = start_process(cmd_dict['arguments'])
        # Write to the standard in #
        if 'stdin' in cmd_dict:
            proc.stdin.write(cmd_dict["stdin"])
            proc.stdin.close()
        # The Future object takes it from here #
        future = Future(proc, cmd_dict)
        # Let's keep a reference of it #
        PARRALEL_JOBS.append(future)
        # Hand it back to the user #
        return future

    def slurm(self, *args, **kwargs):
        """Run a program via the SLURM system and return a Future object."""
        # Optional name #
        name = kwargs.get('name')
        # Define special parameters #
        special_params = (('time','-t'), ('account','-A'), ('name','-J'))
        # Compose the command #
        slurm_cmd = ['srun', '-n', '1', '-Q']
        # Get optional parameters #
        for param,key in special_params:
            if param in kwargs: slurm_cmd += [key, kwargs.pop(param)]
        # Call the user defined function #
        cmd_dict = self.function(*args, **kwargs)
        cmd_dict['arguments'] = [str(a) for a in cmd_dict['arguments']]
        # Get optional keyword parameters #
        qos = kwargs.pop('qos') if 'qos' in kwargs else None
        if qos: slurm_cmd += ['--qos='+qos]
        # Update the command #
        cmd_dict["arguments"] = slurm_cmd + cmd_dict["arguments"]
        # Start a process #
        proc = start_process(cmd_dict['arguments'])
        # Write the standard in #
        if 'stdin' in cmd_dict:
            proc.stdin.write(cmd_dict["stdin"])
            proc.stdin.close()
        # The Future object takes it from here #
        future = Future(proc, cmd_dict, name)
        # Let's keep a reference of it #
        PARRALEL_JOBS.append(future)
        # Hand it back to the user #
        return future

    def lsf(self, *args, **kwargs):
        """Run a program via the LSF system and return a Future object."""
        # Optional name #
        name = kwargs.get('name')
        # Get extra optional keyword parameters #
        queue = kwargs.pop('queue') if 'queue' in kwargs else None
        # Call the user defined function #
        cmd_dict = self.function(*args, **kwargs)
        cmd_dict['arguments'] = [str(a) for a in cmd_dict['arguments']]
        # Compose the command #
        bsub_cmd = ["bsub", "-o", "/dev/null", "-e", "/dev/null", "-K", "-r"]
        if queue: bsub_cmd += ['-q', queue]
        cmd_dict["arguments"] = bsub_cmd + cmd_dict["arguments"]
        # Start a process #
        proc = start_process(cmd_dict['arguments'])
        # Write the standard in #
        if 'stdin' in cmd_dict:
            proc.stdin.write(cmd_dict["stdin"])
            proc.stdin.close()
        # The FutureLSF object takes it from here #
        future = Future(proc, cmd_dict, name)
        # Let's keep a reference of it #
        PARRALEL_JOBS.append(future)
        # Hand it back to the user #
        return future

################################################################################
class Future(object):
    """Object returned when functions decorated with ``@command``
    are executed in parallel with ``parallel()``."""

    def __init__(self, proc, cmd_dict, name=None):
        self.proc = proc
        self.cmd_dict = cmd_dict
        self.name = name

    @property
    def finished(self):
        if self.proc.poll() is None: return False
        else: return True

    @property
    def return_code(self):
        return self.proc.poll()

    def wait(self):
        # Wait until completion #
        try: return_code = self.proc.wait()
        except KeyboardInterrupt as err:
            print "You aborted the process pid %i. It was: %s " % (self.proc.pid, self.cmd_dict["arguments"])
            raise err
        # Read result #
        stdout, stderr = self.proc.stdout.read(), self.proc.stderr.read()
        # Check for failure #
        if return_code != 0: raise CommandFailed(self.cmd_dict["arguments"], stderr, self.name)
        # Return result #
        result = self.cmd_dict.get("return_value")
        if callable(result): return result(stdout, stderr)
        elif result == 'stdout': return stdout
        else: return result


# Built-in modules #
import os, time, getpass

# Internal modules #
from common import split_thousands
from autopaths import FilePath
from collections import OrderedDict

# Third party modules #
import numpy
import matplotlib, brewer2mpl
matplotlib.use('Agg', warn=False)
from matplotlib import pyplot

# Constants #
cool_colors = brewer2mpl.get_map('Set1', 'qualitative', 8).mpl_colors
cool_colors.reverse()
cool_colors += brewer2mpl.get_map('Set2',    'qualitative', 8).mpl_colors
cool_colors += brewer2mpl.get_map('Set3',    'qualitative', 8).mpl_colors
cool_colors += brewer2mpl.get_map('Pastel1', 'qualitative', 8).mpl_colors
cool_colors += brewer2mpl.get_map('Pastel2', 'qualitative', 8).mpl_colors
cool_colors += brewer2mpl.get_map('Greys',   'sequential',  8).mpl_colors

################################################################################
class Graph(object):
    default_params = OrderedDict((
        ('width'  , 12.0),
        ('height' , 7.0),
        ('bottom' , 0.14),
        ('top'    , 0.93),
        ('left'   , 0.06),
        ('right'  , 0.98),
        ('x_grid' , False),
        ('y_grid' , False),
        ('x_scale', None),
        ('y_scale', None),
        ('x_label', None),
        ('y_label', None),
        ('sep'    , ()),
        ('formats', ('pdf',)),
    ))

    def __nonzero__(self): return self.path.__nonzero__()

    def __init__(self, parent=None, base_dir=None, short_name=None):
        # Save parent #
        self.parent = parent
        # Base dir #
        if base_dir is None: self.base_dir = self.parent.p.graphs_dir
        else: self.base_dir = base_dir
        # Short name #
        if short_name: self.short_name = short_name
        if not hasattr(self, 'short_name'): self.short_name = 'graph'
        # Paths #
        self.path = FilePath(self.base_dir + self.short_name + '.pdf')

    def __call__(self, *args, **kwargs):
        """Plot the graph if it doesn't exist. Then return the path to it.
        Force the reruning with rerun=True"""
        if not self or kwargs.get('rerun'): self.plot(*args, **kwargs)
        return self.path

    def save_plot(self, fig, axes, **kwargs):
        # Parameters #
        self.params = {}
        for key in self.default_params:
            if key in kwargs:                          self.params[key] = kwargs[key]
            elif hasattr(self, key):                   self.params[key] = getattr(self, key)
            elif self.default_params[key] is not None: self.params[key] = self.default_params[key]
        # Backwards compatibility #
        if kwargs.get('x_log', False): self.params['x_scale'] = 'symlog'
        if kwargs.get('y_log', False): self.params['y_scale'] = 'symlog'
        # Log #
        if 'x_scale' in self.params: axes.set_xscale(self.params['x_scale'])
        if 'y_scale' in self.params: axes.set_yscale(self.params['y_scale'])
        # Axis limits #
        if 'x_min' in self.params: axes.set_xlim(self.params['x_min'], axes.get_xlim()[1])
        if 'x_max' in self.params: axes.set_xlim(axes.get_xlim()[0], self.params['x_max'])
        if 'y_min' in self.params: axes.set_xlim(self.params['y_min'], axes.get_ylim()[1])
        if 'y_max' in self.params: axes.set_xlim(axes.get_ylim()[0], self.params['y_max'])
        # Title #
        title = self.params.get('title', False)
        if title: axes.set_title(title)
        # Axes labels  #
        if self.params.get('x_label'): axes.set_xlabel(self.params['x_label'])
        if self.params.get('y_label'): axes.set_ylabel(self.params['y_label'])
        # Adjust #
        fig.set_figwidth(self.params['width'])
        fig.set_figheight(self.params['height'])
        fig.subplots_adjust(hspace=0.0, bottom = self.params['bottom'], top   = self.params['top'],
                                        left   = self.params['left'],   right = self.params['right'])
        # Grid #
        axes.xaxis.grid(self.params['x_grid'])
        axes.yaxis.grid(self.params['y_grid'])
        # Data and source extra text #
        if hasattr(self, 'dev_mode') and self.dev_mode is True:
            fig.text(0.99, 0.98, time.asctime(), horizontalalignment='right')
            job_name = os.environ.get('SLURM_JOB_NAME', 'Unnamed')
            user_msg = 'user: %s, job: %s' % (getpass.getuser(), job_name)
            fig.text(0.01, 0.98, user_msg, horizontalalignment='left')
        # Nice digit grouping #
        if 'x' in self.params['sep']:
            separate = lambda x,pos: split_thousands(x)
            axes.xaxis.set_major_formatter(matplotlib.ticker.FuncFormatter(separate))
        if 'y' in self.params['sep']:
            separate = lambda y,pos: split_thousands(y)
            axes.yaxis.set_major_formatter(matplotlib.ticker.FuncFormatter(separate))
        # Add custom labels #
        if 'x_labels' in self.params: axes.set_xticklabels(self.params['x_labels'])
        if 'x_labels_rot' in self.params: pyplot.setp(axes.xaxis.get_majorticklabels(), rotation=self.params['x_labels_rot'])
        # Possibility to overwrite path #
        if 'path' in self.params: path = FilePath(self.params['path'])
        else:                     path = self.path
        # Save it as different formats #
        for ext in self.params['formats']: fig.savefig(path.replace_extension(ext))

    def plot(self, bins=250, **kwargs):
        """An example plot function. You have to subclass this method."""
        # Data #
        counts = [sum(map(len, b.contigs)) for b in self.parent.bins]
        # Linear bins in logarithmic space #
        if 'log' in kwargs.get('x_scale', ''):
            start, stop = numpy.log10(1), numpy.log10(max(counts))
            bins = list(numpy.logspace(start=start, stop=stop, num=bins))
            bins.insert(0, 0)
        # Plot #
        fig = pyplot.figure()
        pyplot.hist(counts, bins=bins, color='gray')
        axes = pyplot.gca()
        # Information #
        title = 'Distribution of the total nucleotide count in the bins'
        axes.set_title(title)
        axes.set_xlabel('Number of nucleotides in a bin')
        axes.set_ylabel('Number of bins with that many nucleotides in them')
        # Save it #
        self.save_plot(fig, axes, **kwargs)
        pyplot.close(fig)
        # For convenience #
        return self

    def save_anim(self, fig, animate, init, bitrate=10000, fps=30):
        """Not functional -- TODO"""
        from matplotlib import animation
        anim = animation.FuncAnimation(fig, animate, init_func=init, frames=360, interval=20)
        FFMpegWriter = animation.writers['ffmpeg']
        writer = FFMpegWriter(bitrate= bitrate, fps=fps)
        # Save #
        self.avi_path = self.base_dir + self.short_name + '.avi'
        anim.save(self.avi_path, writer=writer, codec='x264')



"""
Processes and parallelism
=========================

The ``processes`` module provides some convenience functions
for using processes in python.

Adapted from http://stackoverflow.com/a/16071616/287297

Example usage:

    print prll_map(lambda i: i * 2, [1, 2, 3, 4, 6, 7, 8])

Comments:

"It spawns a predefined amount of workers and only iterates through the input list if there exists an idle worker. I also enabled the "daemon" mode for the workers so that KeyboardInterupt works as expected."

All the stdouts are sent back to the parent stdout, intertwined.
"""

# Modules #
import multiprocessing

################################################################################
def target_func(f, q_in, q_out):
    while True:
        i, x = q_in.get()
        if i is None:
            break
        q_out.put((i, f(x)))

################################################################################
def prll_map(function_to_apply, items, cpus=None):
    # Number of cores #
    if cpus is None: cpus = min(multiprocessing.cpu_count(), 32)
    # Create queues #
    q_in = multiprocessing.Queue(1)
    q_out = multiprocessing.Queue()
    # Process list #
    processes = [multiprocessing.Process(target = target_func,
                                         args   = (function_to_apply, q_in, q_out))
                 for _ in range(cpus)]
    # Start them all #
    for proc in processes:
        proc.daemon = True
        proc.start()
    # Undocumented #
    sent = [q_in.put((i, x)) for i, x in enumerate(items)]
    [q_in.put((None, None)) for _ in range(cpus)]
    res = [q_out.get() for _ in range(len(sent))]
    # Wait for them to finish #
    [proc.join() for proc in processes]
    # Return results #
    return [x for i, x in sorted(res)]


from setuptools import setup

setup(
    name             = 'plumbing',
    version          = '2.0.6',
    description      = 'Helps with plumbing-type programing in python.',
    license          = 'MIT',
    url              = 'http://github.com/xapple/plumbing/',
    author           = 'Lucas Sinclair',
    author_email     = 'lucas.sinclair@me.com',
    classifiers      = ['Topic :: Scientific/Engineering :: Bio-Informatics'],
    packages         = ['plumbing'],
    install_requires = ['sh', 'biopython', 'matplotlib', 'brewer2mpl'],
    long_description = open('README.md').read(),
)



# Built-in modules #
import os, time, inspect
import cPickle as pickle

# Internal modules #
from plumbing.autopaths import FilePath

# Third party modules #
from decorator import decorator

################################################################################
def cached(f):
    """Decorator for functions evaluated only once."""
    def memoized(*args, **kwargs):
        if hasattr(memoized, '__cache__'):
            return memoized.__cache__
        result = f(*args, **kwargs)
        memoized.__cache__ = result
        return result
    return memoized

################################################################################
def property_cached(f):
    """Decorator for properties evaluated only once.
    It can be used to created a cached property like this:

        class Employee(object):
            @property_cached
            def salary(self):
                print "Evaluating..."
                return time.time()
        bob = Employee()
        print bob.salary
        print bob.salary
        bob.salary = "10000$"
        print bob.salary
    """
    # Called when you access the property #
    def retrieve_from_cache(self):
        if '__cache__' not in self.__dict__: self.__cache__ = {}
        if f.__name__ not in self.__cache__:
            if inspect.isgeneratorfunction(f): result = tuple(f(self))
            else: result = f(self)
            self.__cache__[f.__name__] = result
        return self.__cache__[f.__name__]
    # Called when you set the property #
    def overwrite_cache(self, value):
        if '__cache__' not in self.__dict__: self.__cache__ = {}
        self.__cache__[f.__name__] = value
    # Return a wrapper #
    retrieve_from_cache.__doc__ = f.__doc__
    return property(retrieve_from_cache, overwrite_cache)

################################################################################
def property_pickled(f):
    """Same thing as above but the result will be stored on disk
    The path of the pickle file will be determined by looking for the
    `cache_dir` attribute of the instance containing the cached property.
    If no `cache_dir` attribute exists the `p` attribute will be accessed with
    the name of the property being cached."""
    # Called when you access the property #
    def retrieve_from_cache(self):
        # Is it in the cache ? #
        if '__cache__' not in self.__dict__: self.__cache__ = {}
        if f.__name__ in self.__cache__: return self.__cache__[f.__name__]
        # Where should we look in the file system ? #
        if 'cache_dir' in self.__dict__:
            path = FilePath(self.__dict__['cache_dir'] + f.func_name + '.pickle')
        else:
            path = getattr(self.p, f.func_name)
        # Is it on disk ? #
        if path.exists:
            with open(path) as handle: result = pickle.load(handle)
            self.__cache__[f.__name__] = result
            return result
        # Otherwise let's compute it #
        result = f(self)
        with open(path, 'w') as handle: pickle.dump(result, handle)
        self.__cache__[f.__name__] = result
        return result
    # Called when you set the property #
    def overwrite_cache(self, value):
        # Where should we look in the file system ? #
        if 'cache_dir' in self.__dict__:
            path = FilePath(self.__dict__['cache_dir'] + f.func_name + '.pickle')
        else:
            path = getattr(self.p, f.func_name)
        if value is None: path.remove()
        else: raise Exception("You can't set a pickled property, you can only delete it")
    # Return a wrapper #
    retrieve_from_cache.__doc__ = f.__doc__
    return property(retrieve_from_cache, overwrite_cache)

################################################################################
def expiry_every(seconds=0):
    def memoize_with_expiry(func, *args, **kwargs):
        # Get the cache #
        if not hasattr(func, '__cache__'): func.__cache__ = [(0,0)]
        cache = func.__cache__
        # Check the cache #
        if cache:
            result, timestamp = cache[0]
            age = time.time() - timestamp
            if age < seconds: return result
        # Update the cache #
        result = func(*args, **kwargs)
        cache[0] = (result, time.time())
        # Return #
        return result
    return decorator(memoize_with_expiry)

###############################################################################
class LazyString(object):
    """A string-like object that will only compute its value once, when accessed."""
    def __str__(self): return self.value
    def __init__(self, function):
        self._value = None
        self.function = function

    @property
    def value(self):
        if self._value == None: self._value = self.function()
        return self._value

###############################################################################
class LazyDict(object):
    """A dictionary-like object that will only compute its value once, when accessed."""
    def __init__(self, function):
        self._value = None
        self.function = function

    pass

###############################################################################
class class_property(property):
    """You can use this like this:
        class Foo(object):
            @class_property
            @classmethod
            @cached
            def bar(cls):
                return 1+1
    """
    def __get__(self, cls, owner):
        return self.fget.__get__(None, owner)()

